/*-----------------------------------------------------------------*/
/*! 
  \file calcephspkio.c 
  \brief perform I/O on SPICE KERNEL ephemeris data file 
         interpolate SPICE KERNEL Ephemeris data.
         compute position and velocity from SPICE KERNEL ephemeris file.

  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2011, 2012, 2013, 2014, CNRS
   email of the author : gastineau@imcce.fr

  History:                                                                
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
 This file is "triple-licensed", you have to choose one  of the three licenses 
 below to apply on this file.
 
    CeCILL-C
    	The CeCILL-C license is close to the GNU LGPL.
    	( http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html )
   
 or CeCILL-B
        The CeCILL-B license is close to the BSD.
        (http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.txt)
  
 or CeCILL v2.0
      The CeCILL license is compatible with the GNU GPL.
      ( http://www.cecill.info/licences/Licence_CeCILL_V2-en.html )
 

This library is governed by the CeCILL-C, CeCILL-B or the CeCILL license under 
French law and abiding by the rules of distribution of free software.  
You can  use, modify and/ or redistribute the software under the terms 
of the CeCILL-C,CeCILL-B or CeCILL license as circulated by CEA, CNRS and INRIA  
at the following URL "http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C,CeCILL-B or CeCILL license and that you accept its terms.
*/
/*-----------------------------------------------------------------*/

#include "calcephconfig.h"
#if HAVE_STDIO_H
#include <stdio.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#endif
#if HAVE_MATH_H
#include <math.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifndef __cplusplus
#if HAVE_STDBOOL_H
#include <stdbool.h>
#endif
#endif /*__cplusplus */
#if HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif

#include "calcephdebug.h"
#include "real.h"
#define __CALCEPH_WITHIN_CALCEPH 1
#include "calceph.h"
#include "calcephspice.h"
#include "util.h"
#include "calcephinternal.h"

/*--------------------------------------------------------------------------*/
/* private functions */
/*--------------------------------------------------------------------------*/
#if DEBUG
static void calceph_spk_debugsegment(struct SPKSegmentHeader *seg);
#endif
static int calceph_spk_readlistsegment(FILE * file, const char *filename, struct SPKSegmentList **list, int fwd,
                                       int bwd, enum SPKBinaryFileFormat reqconvert);
static int calceph_spk_readsegment_header(FILE * file, const char *filename, struct SPKSegmentHeader *seg);

static void calceph_spk_interpol_PV_an(const double *Coeff_Array, treal TimeJD0,
                                       treal Timediff, stateType * Planet, int N, int ncomp);
static void calceph_spk_interpol_PV_segment_20_an(const double *Coeff_Array, treal TimeJD0,
                                                  treal Timediff, struct SPKSegmentHeader_20 *seg,
                                                  treal Timerecordbeg, stateType * Planet, int N);
static void calceph_spk_convertheader(struct SPKHeader *header, enum SPKBinaryFileFormat reqconvert);

static int calceph_bff_convert_int(int x, enum SPKBinaryFileFormat reqconvert);

static void calceph_bff_convert_array_double(double *x, int n, enum SPKBinaryFileFormat reqconvert);

static int calceph_spk_fastreadword(struct SPKfile *pspk, const struct SPKSegmentHeader *seg, struct SPICEcache *cache,
                                    const char *filename, int rec_begin, int rec_end, const double **record);

/*--------------------------------------------------------------------------*/
/*!  
     Read the header data block for the SPK ephemeris file
     Read the comment block.
     Read the segment block.
     
  @return  1 on sucess and 0 on failure
  
  @param file (inout) file descriptor.   
  @param filename (in) A character string giving the name of an ephemeris data file.   
  @param res (out) descriptor of the ephemeris.  
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_open(FILE * file, const char *filename, struct SPICEkernel *res)
{
    struct SPKHeader header;

    enum SPKBinaryFileFormat reqconvert;

    struct SPKSegmentList *listseg;

    union
    {
        unsigned char c[4];
        unsigned int i;
    } endian;

    listseg = NULL;
 /*---------------------------------------------------------*/
    /* read the header block */
 /*---------------------------------------------------------*/
    /* go to the head of the file */
    if (fseeko(file, 0, SEEK_SET) != 0)
    {
        fatalerror("Can't jump to the beginning of the ephemeris file '%s'\nSystem error : '%s'\n",
                   filename, strerror(errno));
        return 0;
    }
    if (fread(&header, sizeof(header), 1, file) != 1)
    {
        fatalerror("Can't read the beginning of the ephemeris file '%s'\nSystem error : '%s'\n",
                   filename, strerror(errno));
        return 0;
    }

 /*---------------------------------------------------------*/
    /* check if Binary File Format Identification requires a conversion for this hardware */
 /*---------------------------------------------------------*/
    reqconvert = BFF_NATIVE_IEEE;
    endian.i = 0x01020304;
    if (0x04 == endian.c[0])
    {
        /* hardware is little endian */
        if (strncmp("BIG-IEEE", header.bff, BFF_LEN) == 0)
        {
#if DEBUG
            printf("convert 'Binary File Format Identification' to little endian\n");
#endif
            reqconvert = BFF_BIG_IEEE;
        }
    }
    else if (0x01 == endian.c[0])
    {
        /* hardware is big endian */
        if (strncmp("LTL-IEEE", header.bff, BFF_LEN) == 0)
        {
#if DEBUG
            printf("convert 'Binary File Format Identification' to big endian\n");
#endif
            reqconvert = BFF_LTL_IEEE;
        }
    }

    if (reqconvert != BFF_NATIVE_IEEE)
    {
        calceph_spk_convertheader(&header, reqconvert);
    }

#if DEBUG
    calceph_spk_debugheader(&header);
#endif
 /*---------------------------------------------------------*/
    /* check the ftp string */
 /*---------------------------------------------------------*/
    if (calceph_spk_ftp(header.ftp) == 0)
    {
        fatalerror("The FTP string is not valid in the file '%s'\n", filename);
        return 0;
    }
 /*---------------------------------------------------------*/
    /* read the list of segments */
 /*---------------------------------------------------------*/
    if (calceph_spk_readlistsegment(file, filename, &listseg, header.fwd, header.bwd, reqconvert) == 0)
    {
        return 0;
    }
    res->filetype = DAF_SPK;
    res->filedata.spk.header = header;
    res->filedata.spk.list_seg = listseg;
    res->filedata.spk.file = file;
    res->filedata.spk.prefetch = 0;
    res->filedata.spk.mmap_buffer = NULL;
    res->filedata.spk.mmap_size = 0;
    res->filedata.spk.mmap_used = 0;
    return 1;
}

/*--------------------------------------------------------------------------*/
/*! Close the SPICE ephemeris file.
     
  @return  1 on sucess and 0 on failure
  
  @param eph (inout) descriptor of the ephemeris.  
*/
/*--------------------------------------------------------------------------*/
void calceph_spk_close(struct SPKfile *eph)
{
    while (eph->list_seg != NULL)
    {
        struct SPKSegmentList *listseg = eph->list_seg;

        eph->list_seg = eph->list_seg->next;
        free(listseg);
    }
    fclose(eph->file);
    if (eph->mmap_buffer)
    {
#if HAVE_MMAP
        if (eph->mmap_used == 1)
        {
            munmap(eph->mmap_buffer, eph->mmap_size);
        }
        else
#endif
        {
            free(eph->mmap_buffer);
        }
    }
}

/*--------------------------------------------------------------------------*/
/*! Prefetch the SPICE ephemeris file.
     
  @return  1 on sucess and 0 on failure
  
  @param pspk (inout) descriptor of the ephemeris.  
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_prefetch(struct SPKfile *pspk)
{
    int res = 1;

    off_t newlen;

    if (pspk->prefetch == 0)
    {
        if (fseeko(pspk->file, 0, SEEK_END) != 0)
        {
            return 0;
        }
        newlen = ftello(pspk->file);
        if (newlen == -1)
        {
            return 0;
        }

        if (fseeko(pspk->file, 0, SEEK_SET) != 0)
        {
            return 0;
        }

#if HAVE_MMAP
#ifdef MAP_POPULATE
        pspk->mmap_buffer = (double *) mmap(NULL, newlen, PROT_READ, MAP_PRIVATE | MAP_POPULATE, fileno(pspk->file), 0);
#else
        pspk->mmap_buffer = (double *) mmap(NULL, newlen, PROT_READ, MAP_PRIVATE, fileno(pspk->file), 0);
#endif
        if (pspk->mmap_buffer == MAP_FAILED)
        {
            pspk->mmap_buffer = NULL;
            return 0;
        }
        pspk->mmap_size = newlen;
        pspk->mmap_used = 1;
#else
        if (pspk->mmap_size < (unsigned long) newlen)
        {
            pspk->mmap_buffer = (double *) realloc(pspk->mmap_buffer, newlen);
            if (pspk->mmap_buffer == NULL)
            {
                return 0;
            }
            pspk->mmap_size = (unsigned long) newlen;
            pspk->mmap_used = 0;
        }

        if (fread(pspk->mmap_buffer, sizeof(char), newlen, pspk->file) != (size_t) newlen)
        {
            free(pspk->mmap_buffer);
            pspk->mmap_buffer = NULL;
            pspk->mmap_size = 0;
            return 0;
        }
#endif
        pspk->prefetch = 1;
    }
    return res;
}

/*--------------------------------------------------------------------------*/
/*!  
     convert the numbers of the header of a SPK file to the native number 
  
  @param header (inout) header of SPK
  @param reqconvert (in) type of data in the file
*/
/*--------------------------------------------------------------------------*/
static void calceph_spk_convertheader(struct SPKHeader *header, enum SPKBinaryFileFormat reqconvert)
{
    header->nd = calceph_bff_convert_int(header->nd, reqconvert);
    header->ni = calceph_bff_convert_int(header->ni, reqconvert);
    header->fwd = calceph_bff_convert_int(header->fwd, reqconvert);
    header->bwd = calceph_bff_convert_int(header->bwd, reqconvert);
    header->free = calceph_bff_convert_int(header->free, reqconvert);
}

#if DEBUG
/*--------------------------------------------------------------------------*/
/*!  
     Debug the header of a SPK file
  
  @param header (in) header of SPK
*/
/*--------------------------------------------------------------------------*/
void calceph_spk_debugheader(const struct SPKHeader *header)
{
    int k;

    printf("file architecture and type : '%*s'\n", IDWORD_LEN, header->idword);
    printf("ND : %d\n", header->nd);
    printf("NI : %d\n", header->ni);
    printf("Internal file name : '%*s'\n", IFNAME_LEN, header->ifname);
    printf("FWD : %d\n", header->fwd);
    printf("BWD : %d\n", header->bwd);
    printf("FREE : %d\n", header->free);
    printf("Binary file format ID : '%*s'\n", BFF_LEN, header->bff);
    printf("FTP corruption test string : \n");
    for (k = 0; k < FTP_LEN; k++)
        printf("%d ", (int) header->ftp[k]);
    printf("\n");
}
#endif

/*--------------------------------------------------------------------------*/
/*! return 0 if fptbuf is not equal to the string
  FTPSTR:<13>:<10>:<13><10>:<13><0>:<129>:<16><206>:ENDFTP
  
  return 1 if ftp is esqual to this string
  
  @param ftpbuf (in) buffer to checked
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_ftp(const unsigned char ftpbuf[FTP_LEN])
{
    const unsigned char validftp[FTP_LEN] =
        { 'F', 'T', 'P', 'S', 'T', 'R', ':', 13, ':', 10, ':', 13, 10, ':', 13, 0, ':', 129, ':', 16, 206, ':', 'E',
        'N', 'D', 'F', 'T', 'P'
    };
    int j;

    int res = 1;

    for (j = 0; j < FTP_LEN; j++)
    {
        if (ftpbuf[j] != validftp[j])
            res = 0;
    }
    return res;
}

/*--------------------------------------------------------------------------*/
/*! read the list of segments in the file : only the header                 

   @param file (inout) file descriptor.   
   @param filename (in) A character string giving the name of an ephemeris data file.   
   @param list (inout) pointer to the list 
   @param fwd (in) record number to read
   @param bwd (in) previous record number
   @param reqconvert (in) binary file format
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_readlistsegment(FILE * file, const char *filename, struct SPKSegmentList **list, int fwd, int bwd,
                                enum SPKBinaryFileFormat reqconvert)
{
#pragma unused(bwd)
    char descriptor[DAF_RECORD_LEN];

    double *ddescriptor = (double *) descriptor;

    char segmentid[DAF_RECORD_LEN];

    int next = 0, count;

#if DEBUG
    int prev;
#endif
    int res = 1;

    int j, offset;

    struct SPKSegmentList *pnew;

#if DEBUG
    printf("calceph_spk_readlistsegment(%p,%d)\n", *list, fwd);
#endif

    /* go to the head of the file */
    if (fseeko(file, DAF_RECORD_LEN * (fwd - 1), SEEK_SET) != 0)
    {
        fatalerror
            ("Can't jump to the segment descriptor record at location %d of the ephemeris file '%s'\nSystem error : '%s'\n",
             fwd, filename, strerror(errno));
        return 0;
    }
    if (fread(descriptor, sizeof(char), DAF_RECORD_LEN, file) != DAF_RECORD_LEN)
    {
        fatalerror
            ("Can't read the segment descriptor record at location  %d of the ephemeris file '%s'\nSystem error : '%s'\n",
             fwd, filename, strerror(errno));
        return 0;
    }
    if (fread(segmentid, sizeof(char), DAF_RECORD_LEN, file) != DAF_RECORD_LEN)
    {
        fatalerror("Can't read the segment id record at location  %d of the ephemeris file '%s'\nSystem error : '%s'\n",
                   fwd + 1, filename, strerror(errno));
        return 0;
    }

    calceph_bff_convert_array_double(ddescriptor, DAF_RECORD_LEN / 8, reqconvert);

    next = (int) ddescriptor[0];
#if DEBUG
    prev = (int) ddescriptor[1];
#endif
    count = (int) ddescriptor[2];

#if DEBUG
    printf("iprev=%d %d %d\n", prev, next, count);

    for (j = 0; j < DAF_RECORD_LEN / 8; j++)
        printf("%g ", ((double *) descriptor)[j]);
    printf("\n");
    for (j = 0; j < DAF_RECORD_LEN; j++)
        printf("%c ", segmentid[j]);
    printf("\n");
#endif

    /* create a new element of the list */
    pnew = (struct SPKSegmentList *) malloc(sizeof(struct SPKSegmentList));
    if (pnew == NULL)
    {
        fatalerror("Can't allocate memory for a segment descritor\n");
        return 0;
    }
    /* insert in the list */
    pnew->next = *list;
    pnew->prev = NULL;
    if (*list != NULL)
        (*list)->prev = pnew;
    *list = pnew;
    pnew->recordnumber = fwd;
    pnew->array_seg.count = count;

    /* fill the content of the list */
    offset = 2;
    for (j = 0; j < count && res == 1; j++)
    {
        const int *parint = (int *) (ddescriptor + offset + 3);

        struct SPKSegmentHeader *seg = pnew->array_seg.array + j;

        seg->T_begin = ddescriptor[offset + 1];
        seg->T_end = ddescriptor[offset + 2];
        seg->body = parint[0];
        seg->center = parint[1];
        seg->refframe = parint[2];
        seg->datatype = (enum SPKdatatype) parint[3];
        seg->rec_begin = parint[4];
        seg->rec_end = parint[5];
        seg->bff = reqconvert;

        memcpy(seg->id, segmentid + SEGMENTID_LEN * j, SEGMENTID_LEN);
        memcpy(seg->descriptor, ddescriptor + offset, NS_SPK * 8);
#if DEBUG
        printf("before calceph_spk_readsegment_header:\n");
        calceph_spk_debugsegment(seg);
#endif

        res = calceph_spk_readsegment_header(file, filename, seg);
#if DEBUG
        printf("after calceph_spk_readsegment_header:\n");
        calceph_spk_debugsegment(seg);
#endif
        offset += NS_SPK;
    }

    /* go the next records */
    if (next != 0 && res != 0)
    {
        res = calceph_spk_readlistsegment(file, filename, list, next, fwd, reqconvert);
    }
    return res;
}

/*--------------------------------------------------------------------------*/
/*! read the double-precision numbers from rec_begin to rec_end (in number of double-precision numbers)
   @param file (inout) file descriptor.   
   @param filename (in) A character string giving the name of an ephemeris data file.   
   @param rec_begin (inout) first word to read 
   @param rec_end (inout) last word to read 
   @param record (out) array to store the words
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_readword(FILE * file, const char *filename, int rec_begin, int rec_end, double *record)
{
    size_t n = (size_t) (rec_end - rec_begin + 1);

    if (fseeko(file, (rec_begin - 1) * sizeof(double), SEEK_SET) != 0)
    {
        fatalerror("Can't jump to the location %lu of the ephemeris file '%s'\nSystem error : '%s'\n",
                   rec_begin, filename, strerror(errno));
        return 0;
    }
    if (fread(record, sizeof(double), n, file) != n)
    {
        fatalerror("Can't read %d data at location  %d of the ephemeris file '%s'\nSystem error : '%s'\n",
                   n, rec_begin, filename, strerror(errno));
        return 0;
    }

    return 1;
}

/*--------------------------------------------------------------------------*/
/*! prefetch the data for the segment.
   read the double-precision numbers from rec_begin to rec_end (in number of double-precision numbers)
   
   return 0 if error occurs
   return 1 if no error
   
   @param pspk (inout) spk kernel.   
   @param seg (in) segment 
   @param cache (inout) cache of the object
   @param filename (in) A character string giving the name of an ephemeris data file.   
   @param rec_begin (in) first word to read 
   @param rec_end (in) last word to read 
   @param record (out) array to store the words
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_fastreadword(struct SPKfile *pspk, const struct SPKSegmentHeader *seg, struct SPICEcache *cache,
                             const char *filename, int rec_begin, int rec_end, const double **record)
{
    if (pspk->prefetch == 0)
    {
        if (cache->segment != seg || cache->rec_begin != rec_begin)
        {
            size_t newlen = (rec_end - rec_begin + 1) * sizeof(double);

            if (cache->buffer_size < newlen)
            {
                cache->buffer = (double *) realloc(cache->buffer, newlen);
                if (cache->buffer == NULL)
                {
                    fatalerror("Can't allocate memory for %lu bytes.\n", (unsigned long) newlen);
                    return 0;
                }
                cache->buffer_size = newlen;
            }
            if (calceph_spk_readword(pspk->file, filename, rec_begin, rec_end, cache->buffer) == 0)
            {
                return 0;
            }
            cache->rec_begin = rec_begin;
            cache->rec_end = rec_end;
            cache->segment = seg;
        }
        *record = cache->buffer;
    }
    else
    {   /* prefect was done before */
        *record = pspk->mmap_buffer + rec_begin - 1;
    }
    return 1;
}

#if DEBUG
/*--------------------------------------------------------------------------*/
/* ! debug the segment 
   @param seg (inout) segment 
*/
/*--------------------------------------------------------------------------*/
void calceph_spk_debugsegment(struct SPKSegmentHeader *seg)
{
    printf("segment ID : '%.*s'\n", SEGMENTID_LEN, seg->id);
    printf("T_begin : %e -> %e\n", seg->T_begin, seg->T_end);
    printf("body : %d \n", seg->body);
    printf("center : %d \n", seg->center);
    printf("refframe : %d \n", seg->refframe);
    printf("datatype : %d \n", (int) seg->datatype);
    printf("rec_begin : %d \n", seg->rec_begin);
    printf("rec_end : %d \n", seg->rec_end);
    switch (seg->datatype)
    {
        case SPK_SEGTYPE120:
        case SPK_SEGTYPE103:
        case SPK_SEGTYPE102:
        case SPK_SEGTYPE20:
        case SPK_SEGTYPE3:
        case SPK_SEGTYPE2:
            printf("information of the segment : %e %e %d %d\n",
                   seg->seginfo.data2.T_begin, seg->seginfo.data2.T_len,
                   seg->seginfo.data2.count_dataperrecord, seg->seginfo.data2.count_record);
            break;

        default:
            printf("Unsupported segment (type=%d).", (int) seg->datatype);
            break;
    }
}
#endif

/*--------------------------------------------------------------------------*/
/* ! read the data header of this segmentid
   @param file (inout) file descriptor.   
   @param filename (in) A character string giving the name of an ephemeris data file.   
   @param seg (inout) segment 
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_readsegment_header(FILE * file, const char *filename, struct SPKSegmentHeader *seg)
{
    double drecord[12];

    int res = 1;

    switch (seg->datatype)
    {
        case SPK_SEGTYPE103:
        case SPK_SEGTYPE102:
        case SPK_SEGTYPE3:
        case SPK_SEGTYPE2:
            res = calceph_spk_readword(file, filename, seg->rec_end - 3, seg->rec_end, drecord);
            if (res == 1)
            {
                seg->seginfo.data2.T_begin = drecord[0];
                seg->seginfo.data2.T_len = drecord[1];
                seg->seginfo.data2.count_dataperrecord = (int) drecord[2];
                seg->seginfo.data2.count_record = (int) drecord[3];
            }
            break;

        case SPK_SEGTYPE120:
        case SPK_SEGTYPE20:
            res = calceph_spk_readword(file, filename, seg->rec_end - 6, seg->rec_end, drecord);
            if (res == 1)
            {
                seg->seginfo.data20.dscale = drecord[0];
                seg->seginfo.data20.tscale = drecord[1];
                seg->seginfo.data20.T_init_JD = drecord[2];
                seg->seginfo.data20.T_init_FRACTION = drecord[3];
                seg->seginfo.data20.T_begin = seg->seginfo.data20.T_init_JD + seg->seginfo.data20.T_init_FRACTION;
                seg->seginfo.data20.T_len = drecord[4] * seg->seginfo.data20.tscale;
                seg->seginfo.data20.count_dataperrecord = (int) drecord[5];
                seg->seginfo.data20.count_record = (int) drecord[6];
            }
            break;

        default:
            fatalerror("Unsupported segment (type=%d).", (int) seg->datatype);
            res = 0;
            break;
    }
    return res;
}

/*--------------------------------------------------------------------------*/
/*!                                                      
     This function read the data block corresponding to the specified Time
     if this block isn't already in memory.
     
     It computes position and velocity vectors for a selected  
     planetary body from Chebyshev coefficients.          

     return 1 on success
     return 0 on error  
     return -1 if the segment and object is not found
                                                                         
                                                                          
  @param pspk (inout) SPK file  
  @param seg (inout) segment     
  @param cache (inout) cache for the object                   
  @param TimeJD0  (in) Time for which position is desired (Julian Date)
  @param Timediff  (in) offset time from TimeJD0 for which position is desired (Julian Date)
  @param Planet (out) position and veloticity                                                                        
         Planet has position in ua                                                
         Planet has velocity in ua/day
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_interpol_PV_segment(struct SPKfile *pspk, struct SPKSegmentHeader *seg, struct SPICEcache *cache,
                                    treal TimeJD0, treal Timediff, stateType * Planet)
{
    int res = 0;

    /* we found the valid segment */
    switch (seg->datatype)
    {
        case SPK_SEGTYPE103:
        case SPK_SEGTYPE102:
        case SPK_SEGTYPE3:
        case SPK_SEGTYPE2:
            res = calceph_spk_interpol_PV_segment_2(pspk, seg, cache, TimeJD0, Timediff, Planet);
            break;

        case SPK_SEGTYPE120:
        case SPK_SEGTYPE20:
            res = calceph_spk_interpol_PV_segment_20(pspk, seg, cache, TimeJD0, Timediff, Planet);
            break;

        default:
            fatalerror("Unsupported segment (type=%d).\n", (int) seg->datatype);
            break;
    }

#if DEBUG
    if (res)
    {
        int j;

        printf("answer for %23.16E : \n", TimeJD0 + Timediff);
        for (j = 0; j <= 2; j++)
            printf("%g ", Planet->Position[j]);
        puts("\n");
        for (j = 0; j <= 2; j++)
            printf("%g ", Planet->Velocity[j]);
        puts("\n");
    }
#endif
    return res;
}

/*--------------------------------------------------------------------------*/
/*!                                                      
     This function read the data block corresponding to the specified Time
     if this block isn't already in memory.
     
     This function works only over the specified segment.
     The segment must must of type 2, 3, 102 or 103.
     
     It computes position and velocity vectors for a selected  
     planetary body from Chebyshev coefficients.          

                                                                          
  @param pspk (inout) SPK file  
  @param seg (in) segment where is located TimeJD0+timediff 
  @param cache (inout) cache for the object
  @param TimeJD0  (in) Time for which position is desired (Julian Date)
  @param Timediff  (in) offset time from TimeJD0 for which position is desired (Julian Date)
  @param Planet (out) position and veloticity                                                                        
         Planet has position in ua                                                
         Planet has velocity in ua/day
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_interpol_PV_segment_2(struct SPKfile *pspk, struct SPKSegmentHeader *seg, struct SPICEcache *cache,
                                      treal TimeJD0, treal Timediff, stateType * Planet)
{
    const double Timesec = ((TimeJD0 - 2.451545E+06) + Timediff) * 86400E0;

    int recT;                   /* record where is located Timesec */

    const double *drecord;

    int wordT;

    const int nword = seg->seginfo.data2.count_dataperrecord;

#if DEBUG
    int j;
#endif
    int ncomp, N;

    recT = (Timesec - seg->seginfo.data2.T_begin) / seg->seginfo.data2.T_len;
    if (recT == seg->seginfo.data2.count_record &&
        Timesec <= seg->seginfo.data2.T_begin + seg->seginfo.data2.T_len * recT)
        recT = recT - 1;

    if (recT < 0 || recT >= seg->seginfo.data2.count_record)
    {
        fatalerror
            ("Computation of record is not valid for segment of type 2. Looking for time %23.16E. Beginning time in segment : %23.16E\n"
             "Time slice in the segment : %23.16E\n. Number of records: %d\n Coumputed record : %d\n", Timesec,
             seg->seginfo.data2.T_begin, seg->seginfo.data2.T_len, seg->seginfo.data2.count_record, recT);
        return 0;
    }

    /* read the data */
    wordT = (seg->rec_begin + recT * nword);
    if (calceph_spk_fastreadword(pspk, seg, cache, "", wordT, wordT + nword - 1, &drecord) == 0)
    {
        return 0;
    }
    /*for(j=0; j<nword; j++) printf("%23.16E\n", drecord[j]); */
#if DEBUG
    printf
        ("seg->seginfo.data2.T_begin=%23.16E seg->seginfo.data2.T_len=%23.16E begin record=%23.16E end record =%23.16E Timesec=%23.16E\n",
         seg->seginfo.data2.T_begin - seg->seginfo.data2.T_len, seg->seginfo.data2.T_begin + seg->seginfo.data2.T_len,
         drecord[0], drecord[0] + drecord[1], Timesec);
#endif

    /* computation */
    ncomp = (seg->datatype == SPK_SEGTYPE3 || seg->datatype == SPK_SEGTYPE103) ? 6 : 3;
    N = (nword - 2) / ncomp;
    /*printf("ncomp=%d %d\n", ncomp, N); */

    calceph_spk_interpol_PV_an(drecord, (TimeJD0 - 2.451545E+06) * 86400E0, Timediff * 86400E0, Planet, N, ncomp);

#if DEBUG
    for (j = 0; j < 3; j++)
        printf("%23.16E %23.16E\n", Planet->Position[j], Planet->Velocity[j]);
    printf("%23.16E\n",
           sqrt(Planet->Position[0] * Planet->Position[0] + Planet->Position[1] * Planet->Position[1] +
                Planet->Position[2] * Planet->Position[2]));
#endif

    return 1;
}

/*--------------------------------------------------------------------------*/
/*!  This function computes position and velocity vectors                 
     from Chebyshev coefficients from Coeff_Array.                                      
                                          
                                                                          
 @param Coeff_Array (in) array of Tchebychev coefficient                         
 @param TimeJD0  (in) Time for which position is desired (Julian Date)
 @param Timediff  (in) offset time from TimeJD0 for which position is desired (Julian Date)
 @param Planet (out) Pointer to external array to receive the position/velocity.
   Planet has position in ephemerides units                                 
   Planet has velocity in ephemerides units                                 
 @param N(in)    Number of coeff's      
 @param ncomp (in) number of component
*/
/*--------------------------------------------------------------------------*/
static void calceph_spk_interpol_PV_an(const double *Coeff_Array, treal TimeJD0,
                                       treal Timediff, stateType * Planet, int N, int ncomp)
{
    treal Cp[15000], P_Sum, V_Sum, Up[15000], Tc;

    long long int TimeJD0pent = longr(TimeJD0);

    treal TimeJD0frac = TimeJD0 - realr(TimeJD0pent);

    long long int Timediffpent = longr(Timediff);

    treal Timedifffrac = Timediff - realr(Timediffpent);

    long long int T_begpent;

    treal T_begfrac;

    treal deltapent, deltafrac;

    int i, j;

    stateType X;

    const double T_beg = Coeff_Array[0] - Coeff_Array[1];

    const double T_span = Coeff_Array[1] * 2.E0;

    const double *A = Coeff_Array + 2;

  /*--------------------------------------------------------------------------*/
    /* Compute Tc                           */
  /*--------------------------------------------------------------------------*/

    T_begpent = longr(T_beg);
    T_begfrac = T_beg - T_begpent;
    /* code equivalent a :
       Tc = 2.0E0*(Time - T_beg) / T_span - 1.0E0; */
    deltapent = (TimeJD0pent + Timediffpent - T_begpent) / T_span;
    deltafrac = (TimeJD0frac + Timedifffrac - T_begfrac) / T_span;
    Tc = REALC(2.0E0) * (deltapent + deltafrac) - REALC(1.0E0);

  /*--------------------------------------------------------------------------*/
    /* Compute the interpolated position and velocity                           */
  /*--------------------------------------------------------------------------*/

    if (ncomp == 3)
    {
        Cp[0] = REALC(1.0E0);
        Cp[1] = Tc;
        Cp[2] = REALC(2.0E0) * Tc * Tc - REALC(1.E0);

        Up[0] = REALC(0.0E0);
        Up[1] = REALC(1.0E0);
        Up[2] = REALC(4.0E0) * Tc;

        for (j = 3; j < N; j++)
        {
            Cp[j] = REALC(2.0E0) * Tc * Cp[j - 1] - Cp[j - 2];
            Up[j] = REALC(2.0E0) * Tc * Up[j - 1] + REALC(2.0E0) * Cp[j - 1] - Up[j - 2];
        }

        for (i = 0; i < 3; i++) /* Compute interpolating polynomials */
        {

            P_Sum = REALC(0.0E0);   /* Compute interpolated position & velocity */
            V_Sum = REALC(0.0E0);

            for (j = N - 1; j > -1; j--)
                P_Sum = P_Sum + A[j + i * N] * Cp[j];
            for (j = N - 1; j > 0; j--)
                V_Sum = V_Sum + A[j + i * N] * Up[j];

            X.Position[i] = P_Sum;
            X.Velocity[i] = V_Sum * REALC(2.0E0) / (T_span);
        }
    }
    else
    {
        Cp[0] = REALC(1.0E0);
        Cp[1] = Tc;
        Cp[2] = REALC(2.0E0) * Tc * Tc - REALC(1.E0);

        for (j = 3; j < N; j++)
        {
            Cp[j] = REALC(2.0E0) * Tc * Cp[j - 1] - Cp[j - 2];
        }

        for (i = 0; i < 3; i++) /* Compute interpolating polynomials */
        {

            P_Sum = REALC(0.0E0);   /* Compute interpolated position & velocity */
            V_Sum = REALC(0.0E0);

            for (j = N - 1; j > -1; j--)
            {
                P_Sum = P_Sum + A[j + i * N] * Cp[j];
                V_Sum = V_Sum + A[j + (3 + i) * N] * Cp[j];
            }

            X.Position[i] = P_Sum;
            X.Velocity[i] = V_Sum;
        }
    }

  /*--------------------------------------------------------------------------*/
    /*  Return computed values.                                                 */
  /*--------------------------------------------------------------------------*/

    *Planet = X;

}

/*--------------------------------------------------------------------------*/
/*!                                                      
     This function read the data block corresponding to the specified Time
     if this block isn't already in memory.
     
     This function works only over the specified segment.
     The segment must must of type 20 or 120.
     
     It computes position and velocity vectors for a selected  
     planetary body from Chebyshev coefficients.          

                                                                          
  @param pspk (inout) SPK file  
  @param seg (in) segment where is located TimeJD0+timediff 
  @param cache (inout) cache for the object
  @param TimeJD0  (in) Time for which position is desired (Julian Date)
  @param Timediff  (in) offset time from TimeJD0 for which position is desired (Julian Date)
  @param Planet (out) position and veloticity                                                                        
         Planet has position in ua                                                
         Planet has velocity in ua/day
*/
/*--------------------------------------------------------------------------*/
int calceph_spk_interpol_PV_segment_20(struct SPKfile *pspk, struct SPKSegmentHeader *seg, struct SPICEcache *cache,
                                       treal TimeJD0, treal Timediff, stateType * Planet)
{

    int recT;                   /* record where is located Timesec */

    const double *drecord;

    int wordT;

    const int nword = seg->seginfo.data20.count_dataperrecord;

#if DEBUG
    int j;
#endif
    int ncomp, N;

    const double tbegin = (TimeJD0 - seg->seginfo.data20.T_init_JD) + (Timediff - seg->seginfo.data20.T_init_FRACTION);

    const double Timesec = tbegin * 86400E0;

    //recT = (Timesec - tbegin) / seg->seginfo.data20.T_len;
    recT = Timesec / seg->seginfo.data20.T_len;
#if DEBUG
    printf("recT=%d tbegin=%g Timesec=%g\n", recT, tbegin, Timesec);
#endif
    if (recT == seg->seginfo.data20.count_record && Timesec <= seg->seginfo.data20.T_len * recT)
        recT = recT - 1;

    if (recT < 0 || recT >= seg->seginfo.data20.count_record)
    {
        fatalerror
            ("Computation of record is not valid for segment of type 20. Looking for time %23.16E. Beginning time in segment : %23.16E\n"
             "Time slice in the segment : %23.16E\n. Number of records: %d\n Coumputed record : %d\n", Timesec,
             seg->seginfo.data20.T_begin, seg->seginfo.data20.T_len, seg->seginfo.data20.count_record, recT);
        return 0;
    }

    /* read the data */
    wordT = (seg->rec_begin + recT * nword);
    if (calceph_spk_fastreadword(pspk, seg, cache, "", wordT, wordT + nword - 1, &drecord) == 0)
    {
        return 0;
    }
#if  DEBUG
    for (j = 0; j < nword; j++)
        printf("%23.16E\n", drecord[j]);
    printf
        ("seg->seginfo.data20.T_begin=%23.16E seg->seginfo.data2.T_len=%23.16E begin record=%23.16E end record =%23.16E Timesec=%23.16E\n",
         seg->seginfo.data20.T_begin, seg->seginfo.data20.T_len,
         tbegin + seg->seginfo.data20.T_len * recT,
         tbegin + seg->seginfo.data20.T_len * recT + seg->seginfo.data20.T_len, Timesec);
#endif

    /* computation */
    ncomp = 3;
    N = (nword - 2) / ncomp;

    calceph_spk_interpol_PV_segment_20_an(drecord, TimeJD0, Timediff,
                                          &seg->seginfo.data20, seg->seginfo.data20.T_len * recT, Planet, N);

#if DEBUG
    for (j = 0; j < 3; j++)
        printf("%23.16E %23.16E\n", Planet->Position[j], Planet->Velocity[j]);
    printf("%23.16E\n",
           sqrt(Planet->Position[0] * Planet->Position[0] + Planet->Position[1] * Planet->Position[1] +
                Planet->Position[2] * Planet->Position[2]));
#endif

    return 1;
}

/*--------------------------------------------------------------------------*/
/*!  This function computes position and velocity vectors                 
     from Chebyshev coefficients from Coeff_Array.                                      
       Coeff_Array must be compliant with format of type 20                                   
                                                                          
 @param Coeff_Array (in) array of Tchebychev coefficient : only veclocity followed by position at t=0                         
 @param TimeJD0  (in) Time for which position is desired (Julian Date)
 @param Timediff  (in) offset time from TimeJD0 for which position is desired (Julian Date)
 @param seg (in) time ad distance information for the record
 @param Planet (out) Pointer to external array to receive the position/velocity.
   Planet has position in ephemerides units                                 
   Planet has velocity in ephemerides units                                 
 @param N(in)    Number of coeff's      
*/
/*--------------------------------------------------------------------------*/
static void calceph_spk_interpol_PV_segment_20_an(const double *Coeff_Array, treal TimeJD0,
                                                  treal Timediff, struct SPKSegmentHeader_20 *seg,
                                                  treal Timerecordbeg, stateType * Planet, int N)
{
    treal Cp[15000], P_Sum[3], V_Sum[3], Tc, Ip[15000];

    long long int TimeJD0pent = longr(TimeJD0);

    treal TimeJD0frac = TimeJD0 - realr(TimeJD0pent);

    long long int Timediffpent = longr(Timediff);

    treal Timedifffrac = Timediff - realr(Timediffpent);

    long long int T_begpent;

    treal T_begfrac;

    treal deltapent, deltafrac;

    int i, j;

    int flag;

    double d;

    stateType X;

    const double *A = Coeff_Array;

    const treal Dscale = seg->dscale;

    const treal Tscale = seg->tscale;

    const treal T_span = seg->T_len;

  /*--------------------------------------------------------------------------*/
    /* Compute Tc                           */
  /*--------------------------------------------------------------------------*/

    T_begpent = seg->T_init_JD * 86400E0 + Timerecordbeg;
    T_begfrac = seg->T_init_FRACTION * 86400E0;
    /* code equivalent a :
       Tc = 2.0E0*(Time - T_beg) / T_span - 1.0E0; */
    deltapent = ((TimeJD0pent * REALC(86400E0) - T_begpent) + Timediffpent * REALC(86400E0)) / T_span;
    deltafrac = ((TimeJD0frac * REALC(86400E0) - T_begfrac) + Timedifffrac * REALC(86400E0)) / T_span;
    Tc = REALC(2.0E0) * (deltapent + deltafrac) - REALC(1.0E0);

    if (fabs(Tc) > 1.E0)
    {
        printf("deltapent=%23.16E %lld %lld %23.16E\n", TimeJD0pent * 86400E0, T_begpent, Timediffpent, T_span);
        printf("deltafrac=%23.16E %23.16E %23.16E %23.16E\n", TimeJD0frac * 86400E0, T_begfrac, +Timedifffrac, T_span);
        printf("deltapent=%23.16E\n", deltapent);
        printf("deltafrac=%23.16E\n", deltafrac);
    }
#if DEBUG
    printf("Tc=%23.16E %g %g %23.16E\n", Tc, TimeJD0, Timediff, Dscale);
#endif

    /*--------------------------------------------------------------------------*/
    /* Compute the interpolated velocity                           */
    /*--------------------------------------------------------------------------*/
    Cp[0] = REALC(1.0E0);
    Cp[1] = Tc;
    Cp[2] = REALC(2.0E0) * Tc * Tc - REALC(1.E0);

    for (j = 3; j < N; j++)
    {
        Cp[j] = REALC(2.0E0) * Tc * Cp[j - 1] - Cp[j - 2];
    }

    Ip[0] = Tc;
    Ip[1] = (Cp[2] + Cp[0]) / (REALC(4.0E0));

    for (j = 2; j < N - 1; j++)
    {
        Ip[j] = (Cp[j + 1] / (j + 1) - Cp[j - 1] / (j - 1)) / (REALC(2.0E0));
    }

    /* remove constant term from the integral */
    flag = 0;
    d = REALC(2.0E0) * Tc;
    for (i = 3, j = 1; i < N; i += 2, j++)
    {
        d = REALC(0.25E0) / j + REALC(0.25E0) / (j + 1);
        flag = 1 - flag;
        if (flag)
        {
            d = -d;
        }
        Ip[i] += d;
    }

    for (i = 0; i < 3; i++)     /* Compute interpolating polynomials */
    {
#if DEBUG
        printf("utilisation des coef suivants i=%d\n", i);
        for (j = N - 1; j > -1; j--)
            printf("%23.16E\n", A[j + i * (N + 1)]);
        printf("position = %23.16E\n", A[(i + 1) * (N + 1) - 1]);
#endif
        V_Sum[i] = REALC(0.0E0);    /* Compute interpolated velocity */

        for (j = N - 1; j > -1; j--)
        {
            V_Sum[i] = V_Sum[i] + A[j + i * (N + 1)] * Cp[j];
        }
        P_Sum[i] = REALC(0.0E0);    /* Compute interpolated position */
        for (j = N - 2; j > -1; j--)
        {
            P_Sum[i] = P_Sum[i] + A[j + i * (N + 1)] * Ip[j];
        }
        P_Sum[i] = REALC(0.5E0) * P_Sum[i] * (T_span / Tscale) + A[(i + 1) * (N + 1) - 1];

        X.Position[i] = P_Sum[i] * Dscale;
        X.Velocity[i] = V_Sum[i] * Dscale / Tscale;
    }

  /*--------------------------------------------------------------------------*/
    /*  Return computed values.                                                 */
  /*--------------------------------------------------------------------------*/

    *Planet = X;

}

/*--------------------------------------------------------------------------*/
/*!  
     convert a integer to the native number 
  
  @param x (inout) value to convert
  @param reqconvert (in) type of data in the file
*/
/*--------------------------------------------------------------------------*/
static int calceph_bff_convert_int(int x, enum SPKBinaryFileFormat reqconvert)
{
    int r = 0;

    size_t j;

    size_t n = sizeof(int) / sizeof(char);

    switch (reqconvert)
    {
        case BFF_BIG_IEEE:
        case BFF_LTL_IEEE:
            for (j = 0; j < n; j++)
            {
                ((unsigned char *) &r)[j] = ((unsigned char *) &x)[n - 1 - j];
            }
            break;

        default:
            fatalerror("CALCEPH does not handle this conversion format.\n");
            break;
    }
    return r;

}

/*--------------------------------------------------------------------------*/
/*!  
     convert an array of double to the native number 
  
  @param x (inout) array of n "double" 
  @param n (inout) number of floating-points in the array
  @param reqconvert (in) type of data in the file
*/
/*--------------------------------------------------------------------------*/
static void calceph_bff_convert_array_double(double *x, int n, enum SPKBinaryFileFormat reqconvert)
{
    union uswap {
     double x;
     unsigned char r[sizeof(double)];
    }  ur;
    
    int k;

    size_t j;

    size_t w = sizeof(double) / sizeof(char);

    switch (reqconvert)
    {
        case BFF_BIG_IEEE:
        case BFF_LTL_IEEE:
            for (k = 0; k < n; k++)
            {
                for (j = 0; j < w; j++)
                {
                    ur.r[j] = ((unsigned char *) (x + k))[w - 1 - j];
                }
                x[k] = ur.x;
            }
            break;

        case BFF_NATIVE_IEEE:
            break;

        default:
            fatalerror("CALCEPH does not handle this conversion format.\n");
            break;
    }
}
